from typing import Set, Optional
from dbt.node_types import NodeType
from dbt.graph import Graph, UniqueId
from dbt.graph.selector import NodeSelector, SelectionSpec
from dbt.contracts.graph.manifest import Manifest
from dbt.contracts.state import PreviousState


def get_package_names(nodes):
    return set([node.split(".")[1] for node in nodes])


def can_select_indirectly(node):
    """If a node is not selected itself, but its parent(s) are, it may qualify
    for indirect selection.
    Today, only Test nodes can be indirectly selected. In the future,
    other node types or invocation flags might qualify.
    """
    if node.resource_type == NodeType.Test:
        return True
    else:
        return False


class UIDSelector(NodeSelector):
    def __init__(
        self,
        graph: Graph,
        manifest: Manifest,
        previous_state: Optional[PreviousState],
        selected_uid: Set[UniqueId],
        include_empty_nodes: bool = False,
    ) -> None:
        self.selected_uid = selected_uid
        super().__init__(
            graph=graph,
            manifest=manifest,
            previous_state=previous_state,
            include_empty_nodes=include_empty_nodes,
        )

    def get_selected(self, spec: SelectionSpec) -> Set[UniqueId]:
        return self.selected_uid



from idbt.manifest.base import BaseDefinition
from idbt.manifest.i_unit_test import IUnitTestDefinition
from .base import BaseParser
from typing import List
from pydantic import ValidationError
import yaml
from dbt.exceptions import (
    CompilationError,
)

"""
YamlParser read file content and return IUnitTestDefinition
"""


class YamlParser(BaseParser):
    def __init__(self) -> None:
        super().__init__()

    def parse(self, file_path: str) -> List[BaseDefinition]:
        parsed_definitions = []

        try:
            with open(file_path, "r") as file:
                # Load YAML content
                yaml_content: dict = yaml.safe_load(file)
                # parse i_unit_tests
                parsed_definitions = parsed_definitions + self._parse_unit_test_contents(
                    yaml_content, file_path
                )

        except FileNotFoundError:
            raise CompilationError(f"File '{file_path}' not found.")
        except yaml.YAMLError as e:
            raise CompilationError(f"Error parsing YAML file '{file_path}': {e}")
        except ValidationError as e:
            raise CompilationError(e.errors())
        return parsed_definitions

    def _parse_unit_test_contents(
        self, yaml_content: str, file_path: str
    ) -> List[IUnitTestDefinition]:
        result: List[IUnitTestDefinition] = []
        if "i_unit_tests" in yaml_content:
            for test_case in yaml_content["i_unit_tests"]:
                test_case["file_path"] = file_path
                definition = IUnitTestDefinition.model_validate(test_case)
                result.append(definition)
        return result



from abc import ABC, abstractmethod
from idbt.manifest.base import BaseDefinition

"""
parser file content to definition
"""


class BaseParser(ABC):
    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def parse(self) -> BaseDefinition:
        pass



from dataclasses import dataclass, field

from idbt.manifest.base import BaseDefinition
from .i_unit_test import IUnitTestDefinition
from typing import Dict, List

"""
Manifest contains all customized loaded definition
"""


@dataclass
class IdbtManifest:
    i_unit_test: Dict[str, List[IUnitTestDefinition]] = field(default_factory=lambda: {})

    def update(self, definitions: List[BaseDefinition]):
        for definition in definitions:
            if isinstance(definition, IUnitTestDefinition):
                self._update_i_unit_test_definition(definition)

    def _update_i_unit_test_definition(self, definition: IUnitTestDefinition):
        model = definition.model
        if model in self.i_unit_test:
            self.i_unit_test[model].append(definition)
        else:
            self.i_unit_test[model] = [definition]



from dbt.task.seed import SeedTask, SeedRunner
from dbt.exceptions import (
    DbtInternalError,
)
from dbt.contracts.graph.manifest import Manifest
from dbt.graph.queue import GraphQueue
from idbt.context.i_unit_test import IUnitTestContext
from idbt.graph.selector import UIDSelector
from dbt.graph.selector import NodeSelector
from dbt.graph import UniqueId
from dbt.contracts.results import NodeStatus
from typing import AbstractSet, Set
import copy


class IUnitTestSeedRunner(SeedRunner):
    def __init__(self, config, adapter, node, node_index, num_nodes) -> None:
        super().__init__(config, adapter, node, node_index, num_nodes)

    def print_result_line(self, result):
        """
        This hook is to skip model success print
        """
        if result.status == NodeStatus.Success:
            return
        super().print_result_line(result)


"""
This seed task is implemented for selecting specific node's uuid via constructor instead of using NodeSelector
"""


class IUnitTestSeedTask(SeedTask):
    def __init__(
        self,
        args,
        config,
        manifest: Manifest,
        selected_seednodes: Set[UniqueId],
        i_unit_test_context: IUnitTestContext,
    ) -> None:
        self.selected_seednodes = selected_seednodes
        self.i_unit_test_context = i_unit_test_context
        super().__init__(args, config, manifest)

    def get_runner_type(self, _):
        return IUnitTestSeedRunner

    def get_node_selector(self) -> NodeSelector:
        return UIDSelector(
            graph=self.graph,
            manifest=self.manifest,
            previous_state=self.previous_state,
            selected_uid=self.selected_seednodes,
        )

    def before_run(self, adapter, selected_uids: AbstractSet[str]) -> None:
        """
        This hook is to get the required_schemas and assign it into IUnitTestContext
        """
        required_schemas = self.get_model_schemas(adapter, selected_uids)
        for required_schema in required_schemas:
            self.i_unit_test_context.required_schemas.add(required_schema)
        super().before_run(adapter, selected_uids)

    def task_end_messages(self, results) -> None:
        """
        This hook is to skip printing seedtask "Completed successfully"
        """
        pass



from dbt.events.base_types import (
    InfoLevel,
)
from typing import Dict
import colorama
from dbt.events.format import format_fancy_output_line

COLORS: Dict[str, str] = {
    "red": colorama.Fore.RED,
    "green": colorama.Fore.GREEN,
    "yellow": colorama.Fore.YELLOW,
    "blue": colorama.Fore.BLUE,
    "reset_all": colorama.Style.RESET_ALL,
}


def color(text: str, color_code: str) -> str:
    return "{}{}{}".format(color_code, text, COLORS["reset_all"])


def yellow(text: str) -> str:
    return color(text, COLORS["yellow"])


def red(text: str) -> str:
    return color(text, COLORS["red"])


def blue(text: str) -> str:
    return color(text, COLORS["blue"])


class LogModelResult(InfoLevel):
    def code(self) -> str:
        return "IK001"

    def message(self) -> str:

        if self.index == 0 and self.total == 0:
            info = "OK"
            status = yellow("NONE TEST")
        elif self.status == "error":
            info = "ERROR creating"
            status = red("UNIT TEST")
        else:
            info = "OK created"
            status = blue("UNIT TEST")

        msg = f"{info} {self.description}"
        return format_fancy_output_line(
            msg=msg,
            status=status,
            index=self.index,
            total=self.total,
            execution_time=self.execution_time,
        )



from typing import Optional
from dbt.compilation import Compiler
from dbt.contracts.graph.nodes import ManifestSQLNode, Dict, Any
from dbt.contracts.graph.manifest import Manifest
from dbt.config import RuntimeConfig
from dbt.contracts.graph.nodes import (
    ManifestNode,
)
from dbt.context.providers import ModelContext, RuntimeProvider


def generate_runtime_unit_test_context(
    model: ManifestNode,
    config: RuntimeConfig,
    manifest: Manifest,
) -> Dict[str, Any]:

    ctx = ModelContext(model, config, manifest, RuntimeProvider(), None)
    return ctx.to_dict()


class IUnittestCompiler(Compiler):
    def __init__(self, config) -> None:
        super().__init__(config)

    def _compile_code(
        self,
        node: ManifestSQLNode,
        manifest: Manifest,
        extra_context: Optional[Dict[str, Any]] = None,
    ) -> ManifestSQLNode:
        node = super()._compile_code(node, manifest, extra_context)
        return node

    # def _create_node_context(
    #     self,
    #     node: ManifestSQLNode,
    #     manifest: Manifest,
    #     extra_context: Dict[str, Any],
    # ) -> Dict[str, Any]:
    #     return generate_runtime_unit_test_context(node, self.config, manifest)



# N.B.
# This will add to the package’s __path__ all subdirectories of directories on sys.path named after the package which effectively combines both modules into a single namespace (dbt.adapters)
# The matching statement is in plugins/postgres/dbt/__init__.py

from pkgutil import extend_path

__path__ = extend_path(__path__, __name__)


from idbt.exception import FileNotFoundException, ParsingException
from .base import BaseDefinition
from typing import List, Optional, Any
from pydantic import BaseModel, model_validator
import pathlib
import os


class AssertMacroParameter(BaseModel):
    key: str
    value: Any


class Given(BaseModel):
    input: str
    file: str


class AssertMacro(BaseModel):
    name: str
    parameters: List[AssertMacroParameter]


class IUnitTestDefinition(BaseDefinition):
    model: str
    name: str
    description: Optional[str]
    given: List[Given]
    expected_output_file: Optional[str]
    assert_macro: AssertMacro
    file_path: str

    @model_validator(mode="after")
    def post_check(self) -> "IUnitTestDefinition":
        if self.expected_output_file and os.path.exists(self.expected_output_file) == False:
            raise FileNotFoundException(
                f"""expected_output_file: {self.expected_output_file} not found in unitest: {self.file_path}.
                    Please recheck the unittest yaml file {self.file_path}"""
            )

        for given_input in self.given:
            if os.path.exists(given_input.file) == False:
                raise FileNotFoundException(
                    f"""given.file: {given_input.file} not found in unitest: {self.file_path}.
                    Please recheck the unittest yaml file {self.file_path}
                    """
                )

        return self



import copy
from dbt.task.run import ModelRunner
from dbt.contracts.graph.manifest import Manifest
from dbt.parser import manifest
from functools import partialmethod
from dbt.contracts.graph.nodes import ModelNode
from dbt.task.runnable import GraphRunnableTask
from dbt.contracts.graph.nodes import  ManifestNode
   
copied_runtime_initialize = copy.deepcopy(GraphRunnableTask._runtime_initialize)
copied_compile_and_execute =copy.deepcopy(ModelRunner.compile_and_execute)
copied_parse_manifest = copy.deepcopy(manifest.parse_manifest)
copied_execute = copy.deepcopy(ModelRunner.execute)

def _wrapper_runtime_initialize(self,uuid:str):
    """
    this is a wrapper to access the manifest without compiled
    """
    copied_runtime_initialize(self)
    manifest:Manifest = self.manifest
    for attr, value in manifest.nodes.items():
        if(isinstance(value,ModelNode)):
            value.schema = f'{value.schema}_unittest_{uuid}'
          

def _wrapper_compile_and_execute(self, manifest:Manifest, ctx,uuid:str):
    """
    this is a wrapper to access the adapter and do some stuffs before execute each note
    """
    return copied_compile_and_execute(self, manifest, ctx)

def _wrapper_execute(self, model:ManifestNode, manifest:Manifest):
    """
    this is a wrapper to access to process of executing each node
    """
    print('_wrapper_execute')
    
    return copied_execute(self,model,manifest)

def _wrapper_parse_manifest(*args, **kwargs):
    manifest = copied_parse_manifest(*args, **kwargs)
    return manifest

def register(uuid:str):
    """
    Unit test will be run like dbt run flow but change the schema by a uuid prefix.
    This function receive the  schema and register hooks
    Args:
        uuid: uuid string to create schema prefix ex: 5a9ef4bc-3d08-474b-9b02-82faeb146c88
    """
    print('create hook ModelRunner.compile_and_execute')
    # GraphRunnableTask._runtime_initialize = partialmethod(_wrapper_runtime_initialize,uuid=uuid)
    # ModelRunner.compile_and_execute = partialmethod(_wrapper_compile_and_execute,uuid=uuid)
    # ModelRunner.execute = _wrapper_execute
    # manifest.parse_manifest = _wrapper_parse_manifest




    






from idbt.manifest import IdbtManifest
from typing import Dict, Type, List

from idbt.parser.base import BaseParser
from idbt.parser.yaml_parser import YamlParser
import os

"""
FileLoader is used to loop all dbt project folder and load indicated manifest or filepath
"""


class FileLoader:
    parser_classes: Dict[str, Type[BaseParser]] = {"yaml": YamlParser, "yml": YamlParser}
    registered_parsers: Dict[str, BaseParser] = {}

    def __init__(self, file_extensions: List[str] = ["yaml", "yml"]) -> None:
        for file_extension in file_extensions:
            if self.parser_classes.__contains__(file_extension):
                self.registered_parsers[file_extension] = self.parser_classes[file_extension]()

    def load(self, folder_paths: List[str]) -> IdbtManifest:
        """
        Loop folder, get file extension and used registered parser to get definitions
        """
        manifest = IdbtManifest()
        for folder_path in folder_paths:
            for root, _, files in os.walk(folder_path):
                for file in files:
                    file_extension = os.path.splitext(file)[1][1:]
                    if file_extension in self.registered_parsers:
                        parser = self.registered_parsers[file_extension]
                        file_path = os.path.join(root, file)
                        definitions = parser.parse(file_path)
                        manifest.update(definitions)
        return manifest



import uuid
from dbt.task.run import RunTask
import click
from dbt.cli import requires, params as p
from dbt.cli.main import global_flags, cli
from dbt.cli.main import dbtRunner, dbtRunnerResult
from idbt.task.i_unit_test import IUnitTestTask
from idbt.context.i_unit_test import IUnitTestContext
import copy


@cli.command("unittest")
@click.pass_context
@global_flags
@p.defer
@p.deprecated_defer
@p.favor_state
@p.deprecated_favor_state
@p.exclude
@p.full_refresh
@p.profile
@p.profiles_dir
@p.project_dir
@p.select
@p.selector
@p.state
@p.defer_state
@p.deprecated_state
@p.target
@p.target_path
@p.threads
@p.vars
@requires.postflight
@requires.preflight
@requires.profile
@requires.project
@requires.runtime_config
@requires.manifest
def unittest(ctx, **kwargs):
    """Inter-k dbt unittest"""

    uuid_str = str(uuid.uuid4()).replace("-","_")
    source_schema = f"source_unittest_{uuid_str}"
    i_unit_test_context = IUnitTestContext(
        uuid_str,
        source_schema,
        flags=copy.deepcopy(ctx.obj["flags"]),
        config=copy.deepcopy(ctx.obj["runtime_config"]),
        manifest=copy.deepcopy(ctx.obj["manifest"]),
    )

    task = IUnitTestTask(
        ctx.obj["flags"], ctx.obj["runtime_config"], ctx.obj["manifest"], i_unit_test_context
    )

    results = task.run()
    success = task.interpret_results(results)
    return results, success


if __name__ == "__main__":
    cli()



from dataclasses import dataclass, field
from dbt.cli.flags import Flags
from typing import Any, Set
from dbt.config.runtime import RuntimeConfig
from dbt.contracts.graph.manifest import Manifest
from dbt.adapters.base import BaseRelation


@dataclass
class IUnitTestContext:
    schema_suffix: str
    source_schema: str
    flags: Flags
    config: RuntimeConfig
    manifest: Manifest
    required_schemas: Set[BaseRelation] = field(default_factory=lambda: set())



from dataclasses import dataclass
from datetime import datetime
import time
from typing import AbstractSet, List, Optional
from dbt.task.run import RunTask, ModelRunner, fire_event
from dbt.contracts.graph.manifest import Manifest
from dbt.graph import UniqueId
from dbt.contracts.graph.nodes import ModelNode, SeedNode
from idbt.context.i_unit_test import IUnitTestContext
from dbt.config.runtime import RuntimeConfig
from typing import Set, Dict, Union
from idbt.loader import FileLoader
from idbt.manifest import IdbtManifest
from idbt.manifest.i_unit_test import IUnitTestDefinition
from idbt.task.seed import IUnitTestSeedTask
import copy
from dbt.exceptions import (
    DbtRuntimeError,
)
from dbt.contracts.results import NodeStatus, RunResult
from dbt.events.types import (
    LogStartLine,
)
from dbt.events.base_types import EventLevel
from idbt.events.types import LogModelResult as LogIUnitTestResult
from dbt.contracts.results import RunStatus
from dbt.adapters.factory import get_adapter

PARAMETER_PREFIX = "unittest_param"


@dataclass
class TestInputFilesMapper:
    key: str
    file: str
    seednode: SeedNode


class IUnitTestRunner(ModelRunner):
    i_unit_test_context: IUnitTestContext
    idbt_manifest: IdbtManifest
    seednodes: List[SeedNode]

    def __init__(self, config, adapter, node, node_index, num_nodes) -> None:
        super().__init__(config, adapter, node, node_index, num_nodes)

    def compile_and_execute(self, manifest: Manifest, ctx):
        """
        This hook is to load the IDBT manifest before execute each model
        """
        runtime_config: RuntimeConfig = self.config
        self.idbt_manifest = FileLoader().load(runtime_config.model_paths)
        self.seednodes = []
        for key in manifest.nodes:
            node = manifest.nodes[key]
            if isinstance(node, SeedNode):
                self.seednodes.append(node)
        result = super().compile_and_execute(manifest, ctx)
        return result

    def execute(self, model: ModelNode, manifest: Manifest):
        result: RunResult = None
        is_fail_test = False
        if model.identifier in self.idbt_manifest.i_unit_test:
            unit_test_definitions = self.idbt_manifest.i_unit_test[model.identifier]

            for test_index, unit_test_definition in enumerate(unit_test_definitions):
                compiled_model_result, assert_result = self._execute_unit_test(
                    unit_test_definition,
                    manifest,
                    model,
                    test_index,
                    unit_test_definitions.__len__(),
                )
                result = compiled_model_result
                is_fail_test = True if assert_result == False else is_fail_test

        if result == None:
            return RunResult(
                status=RunStatus.Success,
                thread_id="thread_id",
                execution_time=0,
                timing=[],
                message="NOT_FOUND_UNIT_TEST",
                node=model,
                adapter_response={},
                failures=None,
            )
        result.status = NodeStatus.Error if is_fail_test else result.status
        return result

    def print_result_line(self, result: RunResult):
        """
        This hook is to skip model success print and print a warn message for models don't have unittest
        """
        if result.message == "NOT_FOUND_UNIT_TEST":
            fire_event(
                LogIUnitTestResult(
                    description=f"""Not found any unittest for model: {self.node.identifier}""",
                    status=result.status,
                    index=0,
                    total=0,
                    execution_time=result.execution_time,
                    node_info=self.node.node_info,
                ),
                level=EventLevel.INFO,
            )
        if result.status == NodeStatus.Success:
            return
        super().print_result_line(result)

    def describe_node(self):
        """This hook is to change describe of loging"""
        return f"""All unittests for model: {self.node.identifier}"""

    def after_execute(self, result):
        """This hook is to skip handle NodeResult return in execute function if it's None"""
        if result != None:
            super().after_execute(result)

    def _execute_unit_test(
        self,
        unit_test_definition: IUnitTestDefinition,
        manifest: Manifest,
        model: ModelNode,
        test_index: int,
        total_test_case: int,
    ) -> RunResult:
        self._print_unit_test_start_line(unit_test_definition, test_index, total_test_case)
        seednodes, seednode_ids, test_input_file_mappers, expected_output_seednode = (
            self._get_required_seednodes(unit_test_definition)
        )
        self._execute_required_seednodes_tasks(unit_test_definition, seednode_ids, manifest)
        raw_code, additional_context = self._build_jinja_raw_code_and_parameter(
            model, test_input_file_mappers
        )
        recompiled_model = self._recompile_model(model, raw_code, manifest, additional_context)

        with self.adapter.connection_named(
            f"unit_test_{unit_test_definition.model}_{unit_test_definition.name}"
        ):
            compiled_model_result = super().execute(recompiled_model, manifest)

            # DO NOT execute the assert macro if the compiled model run fail#
            if compiled_model_result.status == NodeStatus.Error:
                return compiled_model_result, False

            assert_result = self._execute_assert_macro(
                unit_test_definition, manifest, recompiled_model, expected_output_seednode
            )
            if assert_result == False:
                compiled_model_result.status = NodeStatus.Error
            self._print_unit_test_result_line(
                compiled_model_result, unit_test_definition, test_index, total_test_case
            )
            return compiled_model_result, assert_result

    def _print_unit_test_result_line(
        self,
        result,
        unit_test_definition: IUnitTestDefinition,
        test_index: int,
        total_test_case: int,
    ):
        if result.status == NodeStatus.Error:
            status = result.status
            level = EventLevel.ERROR
        else:
            status = result.message
            level = EventLevel.INFO
        fire_event(
            LogIUnitTestResult(
                description=self._get_unit_test_description(unit_test_definition),
                status=status,
                index=test_index,
                total=total_test_case,
                execution_time=result.execution_time,
                node_info=self.node.node_info,
            ),
            level=level,
        )

    def _get_unit_test_description(self, unit_test_definition: IUnitTestDefinition):
        return f"""Unittest name: {unit_test_definition.name} of model: {self.node.identifier}"""

    def _print_unit_test_start_line(
        self,
        unit_test_definition: IUnitTestDefinition,
        test_index: int,
        total_test_case: int,
    ):
        fire_event(
            LogStartLine(
                description=self._get_unit_test_description(unit_test_definition),
                index=test_index,
                total=total_test_case,
                node_info=self.node.node_info,
            )
        )

    def _execute_assert_macro(
        self,
        unit_test_definition: IUnitTestDefinition,
        manifest: Manifest,
        model: ModelNode,
        expected_output_seednode: SeedNode,
    ) -> bool:
        assert_macro_context = {}
        for param in unit_test_definition.assert_macro.parameters:
            assert_macro_context[param.key] = param.value
        assert_macro_context["model"] = self._get_node_path(model)
        assert_macro_context["expected_output"] = self._get_node_path(expected_output_seednode)

        result = self.adapter.execute_macro(
            unit_test_definition.assert_macro.name,
            kwargs={"assert_macro_context": assert_macro_context},
            manifest=manifest,
        )
        return result

    def _get_node_path(self, node: Union[SeedNode, ModelNode]) -> str:
        relation = self.adapter.Relation.create(
            database=node.database,
            schema=node.schema,
            identifier=node.identifier,
        )
        return str(relation)

    def _recompile_model(
        self,
        old_model: ModelNode,
        raw_code: str,
        manifest: Manifest,
        additional_context: Dict[str, str],
    ) -> ModelNode:
        copied_model = copy.deepcopy(old_model)
        copied_model.raw_code = raw_code
        compiler = self.adapter.get_compiler()
        return compiler.compile_node(copied_model, manifest, additional_context)

    def _build_jinja_raw_code_and_parameter(
        self, model: ModelNode, test_input_file_mappers: List[TestInputFilesMapper]
    ) -> tuple[str, Dict[str, str]]:
        """
        Rebuild the raw_code and additional paramater for re-compiling model node
        Have to use adapter relation to ensure the relation path is suite with dialect convention
        """

        raw_code = model.raw_code
        parameter: Dict[str, str] = {}
        parameter_index = 0

        for test_input_file_mapper in test_input_file_mappers:
            key = f"{PARAMETER_PREFIX}_{parameter_index}"
            raw_code = raw_code.replace(test_input_file_mapper.key, key)
            parameter[key] = self._get_node_path(test_input_file_mapper.seednode)
            parameter_index += 1

        return raw_code, parameter

    def _execute_required_seednodes_tasks(
        self,
        unit_test_definition: IUnitTestDefinition,
        seednode_ids: Set[UniqueId],
        manifest: Manifest,
    ):
        """
        Execute the SeedTask with overiding flags
        Used editted manifest to ensure that the seednodes' schema is changed
        """

        flags = copy.deepcopy(self.i_unit_test_context.flags)
        object.__setattr__(flags, "show", False)
        # EVENT_MANAGER is inialized globally, so reset flag log level won't work
        # flags.__setattr__("QUIET", True)
        # flags.__setattr__("quiet", True)
        # flags.__setattr__("log_level", "error")
        # flags.__setattr__("LOG_LEVEL", "error")
        task = IUnitTestSeedTask(
            flags,
            self.i_unit_test_context.config,
            manifest,
            selected_seednodes=seednode_ids,
            i_unit_test_context=self.i_unit_test_context,
        )
        results = task.run()
        success = task.interpret_results(results)
        if success == False:
            raise DbtRuntimeError(
                f"""Fail uploading seed for unittest.
                Please recheck given input and expected output file for unittest:
                    - model: {unit_test_definition.model}
                    - name:  {unit_test_definition.name}"""
            )

    def _get_required_seednodes(
        self, unit_test_definition: IUnitTestDefinition
    ) -> tuple[List[SeedNode], Set[UniqueId], List[TestInputFilesMapper], SeedNode]:
        """
        Based on test definition, loop all seednodes in manifest to find the correspondence
        Also include the output file seednode
        """

        seednodes: List[SeedNode] = []
        test_input_file_mappers: List[TestInputFilesMapper] = []
        expected_output_seednode: Optional[SeedNode] = None

        for given in unit_test_definition.given:
            file = given.file
            corresponded_seednode = None
            for seednode in self.seednodes:
                if seednode.original_file_path == file:
                    corresponded_seednode = seednode
                    test_input_file_mappers.append(
                        TestInputFilesMapper(
                            key=given.input, file=file, seednode=corresponded_seednode
                        )
                    )
                    seednodes.append(corresponded_seednode)
                    break
            if corresponded_seednode == None:
                raise Exception(f"Not found seednode for file: {file}")

        for seednode in self.seednodes:
            if seednode.original_file_path == unit_test_definition.expected_output_file:
                seednodes.append(seednode)
                expected_output_seednode = seednode
                break

        return (
            seednodes,
            set([node.unique_id for node in seednodes]),
            test_input_file_mappers,
            expected_output_seednode,
        )


class IUnitTestTask(RunTask):
    def __init__(self, args, config, manifest, i_unit_test_context: IUnitTestContext) -> None:
        self.i_unit_test_context = i_unit_test_context
        super().__init__(args, config, manifest)

    def _runtime_initialize(self):
        """
        This hook is to change model bide schema and seed node schema to let the flow create those schema
        """
        super()._runtime_initialize()
        manifest: Manifest = self.manifest
        for attr, value in manifest.nodes.items():
            if isinstance(value, ModelNode):
                value.schema = f"{value.schema}_unittest_{self.i_unit_test_context.schema_suffix}"
            if isinstance(value, SeedNode):
                value.schema = f"seed_data_unittest_{self.i_unit_test_context.schema_suffix}"

    def get_runner_type(self, _):
        return IUnitTestRunner

    def get_runner(self, node) -> IUnitTestRunner:
        runner: IUnitTestRunner = super().get_runner(node)
        runner.i_unit_test_context = self.i_unit_test_context
        return runner

    def before_run(self, adapter, selected_uids: AbstractSet[str]) -> None:
        """
        This hook is to get the required_schemas and assign it into IUnitTestContext
        """
        required_schemas = self.get_model_schemas(adapter, selected_uids)
        for required_schema in required_schemas:
            self.i_unit_test_context.required_schemas.add(required_schema)
        super().before_run(adapter, selected_uids)

    def _delete_unusage_schemas(self, adapter):
        with adapter.connection_named(f"unit_test_delete_unused_schemas"):
            for relation in self.i_unit_test_context.required_schemas:
                adapter.drop_schema(relation)

    def execute_with_hooks(self, selected_uids: AbstractSet[str]):
        """
        This hook is to delete unusage schema in finally block of code
        """
        adapter = get_adapter(self.config)
        self.started_at = time.time()
        try:
            self.before_run(adapter, selected_uids)
            res = self.execute_nodes()
            self.after_run(adapter, res)
        finally:
            self._delete_unusage_schemas(adapter)
            adapter.cleanup_connections()
            elapsed = time.time() - self.started_at
            self.print_results_line(self.node_results, elapsed)
            result = self.get_result(
                results=self.node_results, elapsed_time=elapsed, generated_at=datetime.utcnow()
            )

        return result
